@@ -14,9 +14,10 @@ import os
 import warnings
 
 import tomlkit
+
 from gitarootools.audio.imccontainer import ContainerSubsong, ImcContainer
 from gitarootools.audio.subsong import read_subsong
-from gitarootools.miscutils.extutils import IMCTOML_EXT, SUBSONG_FORMATS
+from gitarootools.miscutils.extutils import SUBSONG_FORMATS
 
 SUBIMC_EXT = SUBSONG_FORMATS["subimc"]
 
@@ -74,25 +75,25 @@ _toml_header = """\
 
 ["Subsong Help/Guide"]
     # This section doesn't do anything. It's just a guide to help you edit the real
-    # Subsong entries below.
-    name = "A_INT"
-    # name: Used internally in Gitaroo Man. Must be 16 ascii characters or less.
+    # [[Subsong]] entries below.
+    name = "SUBSONG"
+    # name: Used internally by the game. Must be 16 ascii characters or less.
     loadmode = "stream"
     # loadmode: Memory loading mode, either "stream" or "entire".
     # "stream" will load only part at a time as it plays, typically used for music.
     # "entire" will load the whole thing in advance, typically used for sound effects.
-    basefile = "00.A_INT.sub.imc"
+    basefile = "00.SUBSONG.sub.imc"
     # basefile: Audio file containing this subsong's audio.
     # Option 1: You can change this to point to another file (such as a WAV file).
     # Option 2: You can use this file as-is and override only specific channels with
     # another file's (see below). If you don't need to modify every channel, Option 2 is
     # recommended because when you recompress a channel, it adds recompression noise and
     # increases the size of the binary diff patch.
-    channels-43-to-56 = "00.A_INT.wav"
+    channels-43-to-56 = "00.OTHER.wav"
     # channels-#-to-#: Use this audio file to override the basefile's channels.
     # Here channels 4 and 3 of this file will override channels 5 and 6 of the basefile
     # in that exact order. (Note: channels are numbered starting from 1, not 0.)
-    channels-to-56 = "00.A_INT.wav"
+    channels-to-56 = "00.OTHER.wav"
     # channels-to-#: Shorthand alternative to the above.
     # The first set of channels will be automatically filled in with channels (1,2,3...)
     # equal to the number of destination channels. In this example, channels 1 and 2 of
@@ -314,12 +315,15 @@ def read_toml(tomlpath):
     return ImcContainer(csubsongs)
 
 
-def write_toml(imccontainer, imcname, outerdestdir, progressfunc=None):
+def write_toml(
+    imccontainer, output_dirpath, output_tomlbase, progressfunc=None,
+):
     """write an ImcContainer to a plaintext toml file and extracted .sub.imc files
 
     imccontainer: ImcContainer instance
-    name: determines the name of the toml dir & toml file, e.g. "ST00A"
-    outerdestdir: directory to which to write the dir containing toml file + subsongs
+    output_dirpath: new directory to create and to unpack IMC contents to
+    output_tomlbase: base filename to which to write the .IMC.toml file (will be put in
+      output_dirpath)
     progressfunc: function to run whenever a subsong is about to be extracted from
       the ImcContainer. It must accept three arguments: an int subsong index,
       an int total number of subsongs, and an imccontainer.ContainerSubsong instance
@@ -333,8 +337,8 @@ def write_toml(imccontainer, imcname, outerdestdir, progressfunc=None):
 
     # prepare toml dir and toml file. writing a bit early here, but if dir/file can't be
     # written, it's better to error before the time-consuming part instead of after
-    tomldir = os.path.join(outerdestdir, imcname)
-    tomlpath = os.path.join(tomldir, f"{imcname}{IMCTOML_EXT}")
+    tomldir = output_dirpath
+    tomlpath = os.path.join(tomldir, output_tomlbase)
     os.makedirs(tomldir, exist_ok=True)
     with open(tomlpath, "wt", encoding="utf-8") as tomlfile:
         try:
