@@ -4,12 +4,15 @@
 
 An IMX image file is a file type from Gitaroo Man that has the extension .IMX
 and contains RGB(A) or indexed image data."""
-from itertools import chain, repeat
+import os
+import pathlib
+from itertools import chain
 from os import SEEK_CUR
 from typing import AnyStr, BinaryIO, Optional, Sequence, Tuple, Union
 
 from PIL import Image
 
+from gitarootools.miscutils.cmdutils import my_warn
 from gitarootools.miscutils.datautils import (
     chunks,
     from_nibbles,
@@ -18,12 +21,12 @@ from gitarootools.miscutils.datautils import (
     readstruct,
     to_nibbles,
     writestruct,
-    zip_to_1st,
 )
 
 SeqIndexed = Sequence[int]
 SeqRGB = Sequence[Tuple[int, int, int]]
 SeqRGBA = Sequence[Tuple[int, int, int, int]]
+BinaryFp = Union[BinaryIO, AnyStr, pathlib.Path]  # binary file or path
 
 PIXEL_FORMATS = ("rgba32", "rgb24", "i8", "i4")
 
@@ -94,27 +97,19 @@ class ImxImage:
                 else:
                     pixfmt = "rgb24"
         self.pixfmt = pixfmt
+        self._alpha128 = alpha128
+        self._palette = palette
+        self._pixels = pixels
 
-        if alpha128:  # use alpha as-is
-            self._pixels = pixels
-            self._palette = palette
+    @property
+    def alpha128(self) -> bool:
+        """True if image uses 128-based alpha"""
+        return self._alpha128
 
-        else:  # convert alpha255 to alpha128
-            if palette is not None:  # convert palette alpha
-                # alpha128 = ceil(alpha255/255*128)
-                self._palette = [
-                    (r, g, b, int(a / 255 * 128 + 0.5)) for (r, g, b, a) in palette
-                ]
-                self._pixels = pixels
-            elif pixels and len(pixels[0]) == 4:  # convert RGBA alpha
-                self._palette = palette
-                # alpha128 = ceil(alpha255/255*128)
-                self._pixels = [
-                    (r, g, b, int(a / 255 * 128 + 0.5)) for (r, g, b, a) in pixels
-                ]
-            else:  # RGB has no alpha
-                self._palette = palette
-                self._pixels = pixels
+    @property
+    def alpha255(self) -> bool:
+        """True if image uses 255-based alpha"""
+        return not self.alpha128
 
     @property
     def size(self) -> Tuple[int, int]:
@@ -132,63 +127,80 @@ class ImxImage:
         return self.haspalette or (self._pixels and len(self._pixels[0]) == 4)
 
     @property
-    def palette(self) -> SeqRGBA:
-        """palette, alpha values are 128-based"""
+    def palette(self) -> Optional[SeqRGBA]:
+        """palette, alpha values are based whether self.alpha128 is True or not"""
         return self._palette
 
+    @property
+    def palette128(self) -> Optional[SeqRGBA]:
+        """palette with 128-based alpha"""
+        if self.palette is None:
+            return None
+        if self.alpha128:
+            return self.palette
+        elif self.alpha255:
+            # get alpha128 = ceil(alpha255/255*128)
+            return [
+                (r, g, b, int(a255 / 255 * 128 + 0.5))
+                for (r, g, b, a255) in self.palette
+            ]
+
+    @property
+    def palette255(self) -> Optional[SeqRGBA]:
+        """palette with 255-based alpha"""
+        if self.palette is None:
+            return None
+        if self.alpha255:
+            return self.palette
+        elif self.alpha128:
+            # get alpha255 = floor(alpha128/128*255)
+            return [
+                (r, g, b, int(a128 / 128 * 255)) for (r, g, b, a128) in self.palette
+            ]
+
     @property
     def pixels(self) -> Union[SeqRGB, SeqRGBA, SeqIndexed]:
-        """pixels, alpha values are 128-based"""
+        """pixels, alpha are based on whether self.alpha128 is true or not"""
         return self._pixels
 
-    def export_palette_rgb_flat(self) -> Optional[SeqRGB]:
-        """return palette RGB as a flat list, for use in exporting to an image format
-
-        Only returns RGB components, concatenated as a flat list. For alpha, use
-        ImxImage.export_palette_alpha_bytes()
-
-        :return None if no palette, or sequence of (r,g,b) colors
-        """
-        palette = self._palette
-        if palette is not None:
-            return tuple(chain.from_iterable(col[:3] for col in palette))
-        else:
-            return None
-
-    def export_palette_alpha_bytes(self) -> Optional[bytes]:
-        """return palette alpha as bytes, for use in exporting to an image format
-
-        :return: None if no palette, or bytes of palette entries' alpha
-        """
-        palette = self._palette
-        if palette is None:
-            return None
-        # convert alpha: alpha255 = floor(alpha128/128*255)
-        alphas255 = (int(col[3] / 128 * 255) for col in palette)
-        # occasional super-opaque values need to be clamped to 255
-        return bytes(a if a <= 255 else 255 for a in alphas255)
-
-    def export_pixels(self) -> Union[SeqRGB, SeqRGBA, SeqIndexed]:
-        """return pixels for use in exporting to an image format
-
-        :return Sequence of indices into the palette if there is a palette, or
-            sequence of (r,g,b) or (r,g,b,a) colors. For rgba, alpha values will have
-            been converted from 128-based alpha to 255-based alpha
-        """
-        pixels = self._pixels
-        if not pixels:
+    @property
+    def pixels128(self) -> Union[SeqRGB, SeqRGBA, SeqIndexed]:
+        """pixels with 128-based alpha"""
+        if not self.pixels:
             return []
-        if self._palette is None and len(pixels[0]) == 4:  # convert RGBA alpha
-            # convert alpha: floor(alpha128/128*255)
-            return [(r, g, b, int(a / 128 * 255)) for (r, g, b, a) in pixels]
-        else:  # don't convert indexed pixels or RGB
-            return pixels
+        if self.palette is None and len(self.pixels[0]) == 4:  # RGBA
+            if self.alpha255:
+                # get alpha128 = ceil(alpha255/255*128)
+                return [
+                    (r, g, b, int(a255 / 255 * 128 + 0.5))
+                    for (r, g, b, a255) in self.pixels
+                ]
+            elif self.alpha128:
+                return self.pixels
+        else:  # RGB or indexed
+            return self.pixels
+
+    @property
+    def pixels255(self) -> Union[SeqRGB, SeqRGBA, SeqIndexed]:
+        """pixels with 255-based alpha"""
+        if not self.pixels:
+            return []
+        if self.palette is None and len(self.pixels[0]) == 4:  # RGBA
+            if self.alpha128:
+                # get alpha255 = floor(alpha128/128*255)
+                return [
+                    (r, g, b, int(a128 / 128 * 255)) for (r, g, b, a128) in self.pixels
+                ]
+            elif self.alpha255:
+                return self.pixels
+        else:  # RGB or indexed
+            return self.pixels
 
 
-def read_imx(file) -> ImxImage:
+def read_imx(fp: BinaryFp) -> ImxImage:
     """read from an IMX image file and return an ImxImage
 
-    :param file: A file path. Or it can be an already-opened file, in which case:
+    :param fp: A file path. Or it can be an already-opened file, in which case:
         - it will read starting from the current file position
         - after returning, file position will be right after the IMX file data
         - the caller is responsible for closing the file afterwards
@@ -196,7 +208,7 @@ def read_imx(file) -> ImxImage:
     :raises EndOfImxImageError if end of IMX data is reached unexpectedly
     :return: ImxImage instance
     """
-    with open_maybe(file, "rb") as file:
+    with open_maybe(fp, "rb") as file:
         try:
             magic = readdata(file, 4)
             if magic != b"IMX\0":
@@ -243,16 +255,16 @@ def read_imx(file) -> ImxImage:
     return ImxImage(width, height, pixels, palette, pixfmt=pixfmt, alpha128=True)
 
 
-def write_imx(imximage: ImxImage, file: Union[BinaryIO, AnyStr]) -> None:
+def write_imx(imximage: ImxImage, fp: BinaryFp) -> None:
     """write imximage to file
 
     :param imximage: an ImxImage instance
-    :param file: A file path. Or it can be an already-opened file, in which case:
+    :param fp: A file path. Or it can be an already-opened file, in which case:
         - it will write starting from the current file position
         - after returning, file position will be right after the written IMX data
         - the caller is responsible for closing the file afterwards
     """
-    with open_maybe(file, "wb") as file:
+    with open_maybe(fp, "wb") as file:
         # header
         file.write(b"IMX\0")
         pixfmtdata = _pixfmt_pixfmtdata[imximage.pixfmt]
@@ -260,14 +272,14 @@ def write_imx(imximage: ImxImage, file: Union[BinaryIO, AnyStr]) -> None:
 
         # palette
         if imximage.pixfmt in ("i4", "i8"):
-            palette = imximage.palette
+            palette = imximage.palette128
             palette_num_bytes = len(palette) * 4
             writestruct(file, "<I", palette_num_bytes)
             writestruct(file, f"{palette_num_bytes}B", *chain.from_iterable(palette))
             writestruct(file, "<I", 2)
 
         # pixels
-        pixels = imximage.pixels
+        pixels = imximage.pixels128
         if imximage.pixfmt == "i8":
             pixels_flat = pixels  # already flat
         elif imximage.pixfmt == "i4":
@@ -282,159 +294,155 @@ def write_imx(imximage: ImxImage, file: Union[BinaryIO, AnyStr]) -> None:
         writestruct(file, "<2I", 3, 0)
 
 
-def read_from_png(
-    file: Union[BinaryIO, AnyStr], pixfmt: Optional[str] = None
-) -> ImxImage:
+def read_from_png(fp: BinaryFp, pixfmt: Optional[str] = None) -> ImxImage:
     """read from an PNG file and return an ImxImage
 
-    :param file: A file object or file path to read from. Only PNG images are guaranteed
+    :param fp: A file object or file path to read from. Only PNG images are guaranteed
         to work, but potentially any image format supported by PIL may also work.
     :param pixfmt: pixel format, one of str (rgba32, rgb24, i8, i4), or None for auto
     :raises ValueError if pixfmt isn't valid
     :raises ValueError if file has a pixel mode unsupported by PIL
     :return: ImxImage instance
     """
-    with Image.open(file) as image:
+    with Image.open(fp) as image:
 
         if pixfmt not in PIXEL_FORMATS and pixfmt is not None:
             raise ValueError(
                 f"pixfmt must be one of {PIXEL_FORMATS} or None, was {pixfmt!r}"
             )
-        if image.mode not in ("RGB", "RGBA", "P"):
-            raise ValueError(f"Image mode {image.mode!r} is not supported yet")
 
         width, height = image.size
 
-        # Choose a pixel format automatically
+        # Choose a pixel format automatically if none is provided
         if pixfmt is None:
-            # potentially convert RGBA to RGB, if lossless
-            if image.mode == "RGBA":
-                image_rgb = image.convert("RGB")
-                is_rgb = tuple(image_rgb.convert("RGBA").getdata()) == tuple(
-                    image.getdata()
-                )
-                if is_rgb:
-                    image = image_rgb
+            num_colors = image.getcolors()
+            num_colors = None if num_colors is None else len(num_colors)
 
-            # potentially convert RGB/RGBA to indexed, if lossless
-            num_colors = len(set(image.getdata()))
-            if image.mode in ("RGBA", "RGB"):
-                if num_colors <= 256:
-                    image = image.quantize(colors=num_colors)
-
-            # determine pixel format
-            if image.mode == "RGBA":
-                pixfmt = "rgba32"
-            elif image.mode == "RGB":
-                pixfmt = "rgb24"
-            elif image.mode == "P" and (num_colors <= 16 and width % 2 == 0):
+            # If few enough colors for i4, do that (but only if width is even)
+            if num_colors is not None and (num_colors <= 16) and (width % 2 == 0):
                 pixfmt = "i4"
-            elif image.mode == "P" and num_colors <= 256:
+
+            # If few enough colors for i8, do that
+            elif image.mode in ("P", "L", "1") or (
+                num_colors is not None and (num_colors <= 256)
+            ):
                 pixfmt = "i8"
+
+            # If no alpha, do RGB. All other cases, do RGBA
             else:
-                raise ValueError(
-                    "Could not automatically choose pixel format, "
-                    f"image.mode={image.mode!r}, num_colors={num_colors!r}"
-                )
+                if image.mode not in ("RGBA", "LA", "PA", "RGBa", "La"):
+                    # if image is in a mode that doesn't have alpha
+                    pixfmt = "rgb24"
+                else:
+                    if image.mode != "RGBA":
+                        image = image.convert("RGBA")
+                    no_alpha = all(a == 255 for a in image.getdata(band=3))
+                    if no_alpha:
+                        pixfmt = "rgb24"
+                    else:
+                        pixfmt = "rgba32"
 
         # get pixels and palette, depending on the desired pixel format
         if pixfmt == "rgba32":
-            image = image.convert("RGBA")
-            pixels = tuple(image.getdata())
+            if image.mode != "RGBA":
+                image = image.convert("RGBA")
+            pixels = image.getdata()
             palette = None
         elif pixfmt == "rgb24":
-            image = image.convert("RGB")
-            pixels = tuple(image.getdata())
+            if image.mode != "RGB":
+                image = image.convert("RGB")
+            pixels = image.getdata()
             palette = None
         elif pixfmt in ("i8", "i4"):
-
-            # Quantize RGB/RGBA image to the right number of colors.
-            # (Do not quantize a "P" (palette) image, convert to "RGBA" first!)
-            if image.mode != "P":
-                image = image.quantize(
-                    colors=(256 if pixfmt == "i8" else 16), dither="FLOYDSTEINBERG"
-                )
-            # If palette image destined for i4 has more than 16 colors, quantize to 16
-            elif pixfmt == "i4" and len(set(image.getdata())) > 16:
-                image = image.convert("RGBA").quantize(
-                    colors=16, dither="FLOYDSTEINBERG"
-                )
-
-            pixels = tuple(image.getdata())
-
-            # Get palette:
-            # 1. get palette alpha from the transparency info. One of 3 possible types:
-            # - iterable of int alpha values corresponding to the palette's rgb values.
-            #   if it's shorter than the palette, rest of the palette is opaque
-            # - single int, the index of the fully transparent palette color
-            # - not present, in which case the entire palette is opaque
-            transparency_info = image.info.get("transparency")
-            if transparency_info is None:
-                # opaque, will be filled in with all 255
-                im_palette_alpha = ()
-            elif not hasattr(transparency_info, "__iter__"):
-                # only one transparent, all the rest is opaque 255
-                transp_idx = transparency_info
-                im_palette_alpha = chain(repeat(255, transp_idx), (0,))
-            else:
-                im_palette_alpha = transparency_info
-            # 2. zip together palette rgb and alpha. If palette alpha is too short, fill
-            # the rest in with 255
-            im_palette_rgb = chunks(image.getpalette(), 3)
-            palette = tuple(
-                (r, g, b, a)
-                for (r, g, b), a in zip_to_1st(
-                    im_palette_rgb, im_palette_alpha, fillvalue=255
-                )
-            )
-
-            # Now we have pixels and palette. If the image is destined for i4, it
-            # already has just 16 colors. But it's likely the palette contains too many
-            # unused entries, so consolidate palette and redo pixels to match.
-            if pixfmt == "i4" and len(palette) > 16:
-                old_palette = palette
-                old_pixels = pixels
-
-                # 1. create new palette containing only used colors,
-                # and keep track of how to get from an old palette index to a new one
-                used_paletteidxs = set(pixels)
-                new_palette = []
-                map_old_new_paletteidxs = dict()
-                newidx = 0
-                for oldidx, oldcol in enumerate(old_palette):
-                    if oldidx in used_paletteidxs:
-                        new_palette.append(oldcol)
-                        map_old_new_paletteidxs[oldidx] = newidx
-                        newidx += 1
-                # 2. redo pixels -- replicate the old pixels using the new palette
-                new_pixels = [map_old_new_paletteidxs[old] for old in old_pixels]
-                # 3. pad palette to 16 colors
-                new_palette = tuple(
-                    chain(new_palette, ((0, 0, 0, 0),) * (16 - len(new_palette)))
-                )
-
-                pixels, palette, = new_pixels, new_palette
+            maxcolors = 256 if pixfmt == "i8" else 16
+            imagename = getimagename(fp)
+            pixels, palette = image2indexed(image, maxcolors, imagename)
 
     return ImxImage(width, height, pixels, palette, pixfmt, alpha128=False)
 
 
-def write_to_png(imximage: ImxImage, file: Union[BinaryIO, AnyStr]) -> None:
+def getimagename(file_or_path: BinaryFp) -> str:
+    """return the base filename from file_or_path
+
+    :param file_or_path: an open file object, string file path, or pathlib Path
+    :return: basename (i.e just filename) of the path (or of the file's path)
+    """
+    if hasattr(file_or_path, "name"):
+        name = str(file_or_path.name)
+    else:
+        name = str(file_or_path)
+    return os.path.basename(name)
+
+
+def image2indexed(
+    image: Image, maxcolors: int = 256, imagename: Optional[str] = None
+) -> Tuple[SeqIndexed, SeqRGBA]:
+    """return (pixels, palette) of image reduced to maxcolors
+
+    Special handling needed here because Pillow's Image.quantize() may reduce colors
+    excessively, resulting in poor image quality
+    (https://github.com/python-pillow/Pillow/issues/5204)
+
+    :param image: Pillow Image instance
+    :param maxcolors: maximum number of colors allowed in the returned palette
+    :param imagename: optional name of image, for printing warnings of excessive color
+        reduction. If None, do not print any warnings.
+    :return: (pixels, palette) where pixels are indices into palette, palette is RGBA
+        colors
+    """
+    if image.mode != "RGBA":
+        image = image.convert("RGBA")
+    imgcolors = image.getcolors(maxcolors=maxcolors)
+
+    # if colors exceed maxcolors, use Pillow's built-in quantization to reduce colors.
+    if imgcolors is None:  # i.e. if getcolors() exceeded maxcolors
+
+        # Sometimes the same image will quantize better from RGB mode than from RGBA
+        # (see https://github.com/python-pillow/Pillow/issues/5217)
+        # Therefore, convert to RGB before quantization if there is no alpha
+        no_alpha = all(a == 255 for a in image.getdata(band=3))
+        if no_alpha:
+            image = image.convert("RGB")
+        image = image.quantize(colors=maxcolors, dither="FLOYDSTEINBERG")
+
+        # warn if there was excessive color reduction (due to a possible Pillow bug)
+        num_destcolors = len(image.getcolors(maxcolors=maxcolors))
+        if imagename is not None and (num_destcolors < maxcolors):
+            my_warn(
+                f"WARNING: {imagename} reduced to {num_destcolors} colors instead of "
+                f"{maxcolors} (due to a possible Pillow bug). To avoid this, manually "
+                f"reduce your image to {maxcolors} or fewer colors before converting."
+            )
+        image = image.convert("RGBA")
+        imgcolors = image.getcolors(maxcolors=maxcolors)
+
+    # now that colors are reduced sufficiently, do manual pixel/palette extraction
+    # (palette is sorted from most to least used colors)
+    palette = [color for count, color in sorted(imgcolors, reverse=True)]
+    color2pixel = {color: idx for idx, color in enumerate(palette)}
+    pixels = [color2pixel[color] for color in image.getdata()]
+    return pixels, palette
+
+
+def write_to_png(imximage: ImxImage, fp: BinaryFp) -> None:
     """write an ImxImage to PNG image file
 
     :param imximage: an ImxImage instance
-    :param file: A file path or an already-opened file
+    :param fp: A file path or an already-opened file
     """
     if imximage.haspalette:
-        image = Image.new("P", imximage.size)
-        image.putpalette(imximage.export_palette_rgb_flat())
-        image.info["transparency"] = imximage.export_palette_alpha_bytes()
+        image = Image.new("RGBA", imximage.size)
+        imx_palette = imximage.palette255
+        imx_pixels_rgba = [imx_palette[i] for i in imximage.pixels]
+        image.putdata(imx_pixels_rgba)
     elif imximage.hasalpha:
         image = Image.new("RGBA", imximage.size)
+        image.putdata(imximage.pixels255)
     else:
         image = Image.new("RGB", imximage.size)
-    image.putdata(imximage.export_pixels())
+        image.putdata(imximage.pixels255)
 
-    image.save(file, format="png")
+    image.save(fp, format="png")
 
 
 def fast_imx_pixfmt(file_or_data: Union[AnyStr, BinaryIO, bytes]) -> str:
